﻿Imports System.Diagnostics
Imports System.IO

Module Module1

    Sub Main()
        Dim numbers() As Integer = {34, 7, 23, 32, 5, 62, 99, 12, 1, 8, 56, 45, 67, 78, 90, 3, 6, 4, 10, 11}

        Console.WriteLine("Original array: " & String.Join(", ", numbers))

        Console.WriteLine("Choose sorting option:")
        Console.WriteLine("1. Manual sort selection")
        Console.WriteLine("2. Benchmark all methods")
        Console.Write("Enter your choice (1 or 2): ")
        Dim modeChoice As String = Console.ReadLine()

        If modeChoice = "1" Then
            ManualSort(numbers)
        ElseIf modeChoice = "2" Then
            BenchmarkSorts(numbers)
        Else
            Console.WriteLine("Invalid choice.")
        End If

        Console.ReadLine()
    End Sub

    ' --- Benchmark Mode ---
    Sub BenchmarkSorts(original() As Integer)
        Dim methods As New Dictionary(Of String, Func(Of Integer(), Integer())) From {
            {"QuickSort", Function(arr)
                              Dim a = CType(arr.Clone(), Integer())
                              FastSort(a, 0, a.Length - 1)
                              Return a
                          End Function},
            {"BubbleSort", Function(arr)
                               Dim a = CType(arr.Clone(), Integer())
                               BubbleSort(a)
                               Return a
                           End Function},
            {"MergeSort", Function(arr) MergeSort(CType(arr.Clone(), Integer()))},
            {"InsertionSort", Function(arr)
                                  Dim a = CType(arr.Clone(), Integer())
                                  InsertionSort(a)
                                  Return a
                              End Function},
            {"SelectionSort", Function(arr)
                                  Dim a = CType(arr.Clone(), Integer())
                                  SelectionSort(a)
                                  Return a
                              End Function}
        }

        Dim results As New Dictionary(Of String, Tuple(Of Integer(), Long))
                                  Dim stopwatch As New Stopwatch()

                                  Console.WriteLine(vbCrLf & "--- Benchmarking Sorts ---")
                                  For Each method In methods
                                      stopwatch.Restart()
                                      Dim sorted = method.Value(original)
                                      stopwatch.Stop()

                                      results(method.Key) = Tuple.Create(sorted, stopwatch.ElapsedMilliseconds)
                                      Console.WriteLine($"{method.Key,-15} Time: {stopwatch.ElapsedMilliseconds} ms")
                                  Next

                                  ' Verify all arrays are equal
                                  Dim reference = String.Join(",", results.First().Value.Item1)
                                  Dim allMatch = results.All(Function(r) String.Join(",", r.Value.Item1) = reference)
                                  Console.WriteLine(vbCrLf & If(allMatch, "✅ All sorting results match.", "❌ Sorting results differ!"))

                                  ' Show one result
                                  Console.WriteLine(vbCrLf & "Sample sorted output (QuickSort):")
                                  Console.WriteLine(String.Join(", ", results("QuickSort").Item1))
    End Sub

    ' --- Manual Sort Mode ---
    Sub ManualSort(numbers() As Integer)
        Console.WriteLine("Choose sorting method:")
        Console.WriteLine("1. QuickSort (Fast)")
        Console.WriteLine("2. BubbleSort (Simple)")
        Console.WriteLine("3. MergeSort (Stable)")
        Console.WriteLine("4. InsertionSort (Good for small arrays)")
        Console.WriteLine("5. SelectionSort (Simple, educational)")
        Console.Write("Enter your choice (1–5): ")
        Dim sortChoice As String = Console.ReadLine()

        Console.Write("Sort order (A = Ascending, D = Descending): ")
        Dim orderChoice As String = Console.ReadLine().ToUpper()

        Dim arrToSort() As Integer = CType(numbers.Clone(), Integer())
        Dim stopwatch As New Stopwatch()
        stopwatch.Start()

        Select Case sortChoice
            Case "1"
                FastSort(arrToSort, 0, arrToSort.Length - 1)
            Case "2"
                BubbleSort(arrToSort)
            Case "3"
                arrToSort = MergeSort(arrToSort)
            Case "4"
                InsertionSort(arrToSort)
            Case "5"
                SelectionSort(arrToSort)
            Case Else
                Console.WriteLine("Invalid sorting method.")
                Return
        End Select
        stopwatch.Stop()

        If orderChoice = "D" Then
            Array.Reverse(arrToSort)
        End If

        Console.WriteLine("Sorted array: " & String.Join(", ", arrToSort))
        Console.WriteLine("Execution Time: " & stopwatch.ElapsedMilliseconds & " ms")

        Console.Write("Export result to file? (Y/N): ")
        Dim exportChoice As String = Console.ReadLine().ToUpper()
        If exportChoice = "Y" Then
            Dim filePath As String = "sorted_output.txt"
            File.WriteAllText(filePath, "Sorted array: " & String.Join(", ", arrToSort))
            Console.WriteLine("Result saved to: " & Path.GetFullPath(filePath))
        End If
    End Sub

    ' --- Sorting Algorithms ---
    Sub FastSort(ByRef arr() As Integer, ByVal low As Integer, ByVal high As Integer)
        If low < high Then
            Dim pivotIndex As Integer = Partition(arr, low, high)
            FastSort(arr, low, pivotIndex - 1)
            FastSort(arr, pivotIndex + 1, high)
        End If
    End Sub

    Function Partition(ByRef arr() As Integer, ByVal low As Integer, ByVal high As Integer) As Integer
        Dim pivot As Integer = arr(high)
        Dim i As Integer = low - 1
        For j As Integer = low To high - 1
            If arr(j) <= pivot Then
                i += 1
                Swap(arr(i), arr(j))
            End If
        Next
        Swap(arr(i + 1), arr(high))
        Return i + 1
    End Function

    Sub BubbleSort(ByRef arr() As Integer)
        Dim n As Integer = arr.Length
        For i As Integer = 0 To n - 2
            For j As Integer = 0 To n - i - 2
                If arr(j) > arr(j + 1) Then
                    Swap(arr(j), arr(j + 1))
                End If
            Next
        Next
    End Sub

    Function MergeSort(arr() As Integer) As Integer()
        If arr.Length <= 1 Then Return arr
        Dim mid As Integer = arr.Length \ 2
        Dim left() As Integer = arr.Take(mid).ToArray()
        Dim right() As Integer = arr.Skip(mid).ToArray()
        Return Merge(MergeSort(left), MergeSort(right))
    End Function

    Function Merge(left() As Integer, right() As Integer) As Integer()
        Dim result As New List(Of Integer)
        Dim i As Integer = 0, j As Integer = 0
        While i < left.Length AndAlso j < right.Length
            If left(i) <= right(j) Then
                result.Add(left(i))
                i += 1
            Else
                result.Add(right(j))
                j += 1
            End If
        End While
        While i < left.Length
            result.Add(left(i)) : i += 1
        End While
        While j < right.Length
            result.Add(right(j)) : j += 1
        End While
        Return result.ToArray()
    End Function

    Sub InsertionSort(ByRef arr() As Integer)
        For i As Integer = 1 To arr.Length - 1
            Dim key = arr(i)
            Dim j = i - 1
            While j >= 0 AndAlso arr(j) > key
                arr(j + 1) = arr(j)
                j -= 1
            End While
            arr(j + 1) = key
        Next
    End Sub

    Sub Swap(ByRef a As Integer, ByRef b As Integer)
        Dim temp = a
        a = b
        b = temp
    End Sub

    Sub SelectionSort(ByRef arr() As Integer)
        Dim n As Integer = arr.Length
        For i As Integer = 0 To n - 2
            Dim minIndex As Integer = i
            For j As Integer = i + 1 To n - 1
                If arr(j) < arr(minIndex) Then
                    minIndex = j
                End If
            Next
            If minIndex <> i Then
                Swap(arr(i), arr(minIndex))
            End If
        Next
    End Sub

End Module

